# -*- python -*-
# $RCSfile: fieldinit.py,v $
# $Revision: 1.37 $
# $Author: rlua $
# $Date: 2007-10-19 20:01:41 $


# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from ooflib.SWIG.engine import field
from ooflib.common import debug
from ooflib.common import registeredclass
from ooflib.common import strfunction
from ooflib.common.IO import parameter
from ooflib.common.IO import whoville
from ooflib.common.IO import xmlmenudump
import ooflib.engine.mesh
import string
import struct
import types

# FieldInit subclasses must be derived from FieldInit *and* from
# RegisteredClass.

class FieldInit:
    def apply(self, femesh, field, singleFieldDef=0):
        if singleFieldDef:
            #The purpose of calling node.fieldDefCount is to
            #make sure we don't set the value of a field
            #that is defined in more than one subproblem
            #when that field gets defined on the subproblem.
            fniter = femesh.funcnode_iterator()
            while not fniter.end():
                node = fniter.node()
                if node.fieldDefCount(field)==1:
                    position = node.position()
                    for i in range(field.ndof()):
                        field.setvalue(node, i, self.func(position, i))
                fniter.next()
        else:
            fniter = femesh.funcnode_iterator()
            while not fniter.end():
                node = fniter.node()
                if node.hasField(field):
                    position = node.position()
                    for i in range(field.ndof()):
                        field.setvalue(node, i, self.func(position, i))
                fniter.next()

# The FieldInitParameter is not a simple RegisteredParameter because
# it has to handle more than one RegisteredClass, for different
# varieties of Fields. 

class FieldInitParameter(parameter.Parameter):
    types = (FieldInit,)
    def binaryRepr(self, datafile, value):
        return value.binaryRepr(datafile) # uses RegisteredClass.binaryRepr
    def binaryRead(self, parser):
        (regkey,) = struct.unpack('>i', parser.getBytes(struct.calcsize('>i')))
        registration = parser.getObject(regkey)
        argdict = {}
        for param in registration.params:
            argdict[param.name] = param.binaryRead(parser)
        return registration(**argdict)
    def valueDesc(self):
        return """A field initializer from one of the <link
        linkend='RegisteredClass:ScalarFieldInit'><classname>ScalarFieldInit</classname></link>
        or <link
        linkend='RegisteredClass:TwoVectorFieldInit'><classname>TwoVectorFieldInit</classname></link>
        classes."""

#############################

# Derived classes of FieldInit must provide func(position, component)
# that returns the value of a Field component at the given position.
# They must also indicate what type of field they work for by
# inserting themselves into the fieldInitDict dictionary.

fieldInitDict = {}

class ScalarFieldInit(registeredclass.RegisteredClass, FieldInit):
    registry = []
    tip="Methods for initializing a scalar field."
    discussion=xmlmenudump.loadFile(
        'DISCUSSIONS/engine/reg/scalar_field_init.xml')
fieldInitDict['ScalarField'] = ScalarFieldInit
fieldInitDict['ScalarFieldBase'] = ScalarFieldInit

class TwoVectorFieldInit(registeredclass.RegisteredClass, FieldInit):
    registry = []
    tip="Methods for initializing a two dimensional vector field."
    discussion=xmlmenudump.loadFile(
        'DISCUSSIONS/engine/reg/two_vector_field_init.xml')
fieldInitDict['TwoVectorField'] = TwoVectorFieldInit


class ThreeVectorFieldInit(registeredclass.RegisteredClass, FieldInit):
    registry = []
    discussion=xmlmenudump.loadFile(
        'DISCUSSIONS/engine/reg/three_vector_field_init.xml')

fieldInitDict['ThreeVectorField'] = ThreeVectorFieldInit

# Subclasses of this are stubs, existing mainly to populate the
# dictionary -- this class does everything.
class OtherMeshFieldInit:
    def __init__(self, mesh):
        self.mesh = mesh                # name of the other mesh
        self.meshobj = ooflib.engine.mesh.meshes[self.mesh]
        self.source_femesh = self.meshobj.getObject()
    def apply(self, femesh, field):
        femesh.init_field(self.meshobj.skeleton, self.source_femesh, field)
    def shortrepr(self):
        return "mesh=%s" % self.mesh

# NVectorField isn't swigged yet
##class NVectorFieldInit(registeredclass.RegisteredClass, FieldInit):
##    registry = []
##fieldInitDict[field.NVectorField] = NVectorFieldInit
   


################    

class ConstScalarFieldInit(ScalarFieldInit):
    def __init__(self, value):
        self.value = value
    def func(self, position, component):
        return self.value
    def shortrepr(self):
        return str(self.value)

registeredclass.Registration(
    "Constant",
    ScalarFieldInit,
    ConstScalarFieldInit,
    0,
    params= [parameter.FloatParameter('value', 0.0,
                                      tip='Value to assign to the Field.')],
    tip="Initialize a scalar field with a constant value.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/const_scalar_field_init.xml')
    )

################

class ConstTwoVectorFieldInit(TwoVectorFieldInit):
    def __init__(self, cx, cy):
        self.cx = cx
        self.cy = cy
    def func(self, position, component):
        if component == 0:
            return self.cx
        return self.cy
    def shortrepr(self):
        return "cx=%s, cy=%s" % (str(self.cx), str(self.cy))

registeredclass.Registration(
    "Constant",
    TwoVectorFieldInit,
    ConstTwoVectorFieldInit,
    0,
    params=[
    parameter.FloatParameter('cx', 0.0,
                      tip='Value to assign to the x component of the Field.'),
    parameter.FloatParameter('cy', 0.0,
                      tip='Value to assign to the y component of the Field.')
    ],
    tip="Initialize a two-vector field with a constant value.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/const_twovec_init.xml')
    )

################

class ConstThreeVectorFieldInit(ThreeVectorFieldInit):
    def __init__(self, cx, cy, cz):
        # Because this is a registered class, it must have attributes
        # corresponding to its paramter names, so we can't be clever
        # with the storage scheme here.
        self.cx = cx
        self.cy = cy
        self.cz = cz
    def func(self, position, component):
        if component==0:
            return self.cx
        else:
            if component==1:
                return self.cy
            else:
                return self.cz
    def shortrepr(self):
        return "cx=%s, cy=%s, cz=%s" % (self.cx, self.cy, self.cz)

registeredclass.Registration(
    "Constant",
    ThreeVectorFieldInit,
    ConstThreeVectorFieldInit,
    0,
    params=[
    parameter.FloatParameter('cx', 0.0,
                   tip='Value to assign to the x component of the Field.'),
    parameter.FloatParameter('cy', 0.0,
                   tip='Value to assign to the y component of the Field.'),
    parameter.FloatParameter('cz', 0.0,
                   tip='Value to assign to the z component of the Field.') 
    ],
    tip="Initialize a two-vector field with a constant value.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/const_threevec_init.xml')
    )

################

class FuncScalarFieldInit(ScalarFieldInit):
    def __init__(self, function):
        self.function = function
    def func(self, position, component):
        return self.function(position.x, position.y)
    def shortrepr(self):
        return str(self.function)

registeredclass.Registration(
    "XYFunction",
    ScalarFieldInit,
    FuncScalarFieldInit,
    1,
    params=[
    strfunction.XYStrFunctionParameter('function', default='0.0',
                                       tip='A function of x and y')],
    tip="Initialize a scalar field with a function of x and y.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/func_scalar_init.xml')
    )

################

class FuncTwoVectorFieldInit(TwoVectorFieldInit):
    def __init__(self, fx, fy):
        self.fx = fx
        self.fy = fy
    def func(self, position, component):
        if component == 0:
            return self.fx(position.x, position.y)
        return self.fy(position.x, position.y)
    def shortrepr(self):
        return "fx=%s, fy=%s" % (str(self.fx), str(self.fy))


registeredclass.Registration(
    "XYFunction",
    TwoVectorFieldInit,
    FuncTwoVectorFieldInit,
    1,
    params=[
    strfunction.XYStrFunctionParameter('fx', value='0.0',
                   tip="The x component of the field as a function of x and y"),
    strfunction.XYStrFunctionParameter('fy', value='0.0',
                   tip="The y component of the field as a function of x and y"),
    ],
    tip="Initialize a two-vector field with a function of x and y.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/func_twovec_init.xml')
    )
                             
##################


class FuncThreeVectorFieldInit(ThreeVectorFieldInit):
    def __init__(self, fx, fy, fz):
        # Registered classes should have attributes corresponding to
        # their parameter names, otherwise these would be in a list.
        self.fx = fx
        self.fy = fy
        self.fz = fz
    def func(self, position, component):
        if component==0:
            return self.fx(position.x, position.y)
        else:
            if component==1:
                return self.fy(postiion.x, position.y)
            else:
                return self.fz(position.x, position.y)
    def shortrepr(self):
        return "fx=%s, fy=%s, fz=%s" % (self.fx, self.fy, self.fz)


registeredclass.Registration(
    "XYFunction",
    ThreeVectorFieldInit,
    FuncThreeVectorFieldInit,
    1,
    params=[
    strfunction.XYStrFunctionParameter('fx', value='0.0',
                   tip="The x component of the field as a function of x and y"),
    strfunction.XYStrFunctionParameter('fy', value='0.0',
                   tip="The y component of the field as a function of x and y"),
    strfunction.XYStrFunctionParameter('fz', value='0.0',
                   tip="The z component of the field as a function of x and y")
    ],
    tip="Initialize a three-vector field with a function of x and y.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/func_threevec_init.xml')
    ) 

    
##################

class ScalarOtherMeshInit(OtherMeshFieldInit, ScalarFieldInit):
    pass
    
registeredclass.Registration(
    "Other Mesh",
    ScalarFieldInit,
    ScalarOtherMeshInit,
    2,
    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                  tip='Copy field values from this Mesh.')],
    tip="Initialize a scalar field from the same field on another mesh.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/scalar_othermesh_init.xml')
    )

class TwoVectorOtherMeshInit(OtherMeshFieldInit, TwoVectorFieldInit):
    pass
            
registeredclass.Registration(
    "Other Mesh",
    TwoVectorFieldInit,
    TwoVectorOtherMeshInit,
    2,
    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                  tip='Copy field values from this Mesh.')],
    tip="Initialize a two-vector field from the same field on another mesh.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/twovec_othermesh_init.xml')
    )

class ThreeVectorOtherMeshInit(OtherMeshFieldInit, ThreeVectorFieldInit):
    pass

registeredclass.Registration(
    "Other Mesh",
    ThreeVectorFieldInit,
    ThreeVectorOtherMeshInit,
    2,
    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                  tip='Copy field values from this Mesh.')],
    tip="Initialize a three-vector field from the same field on another mesh.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/threevec_othermesh_init.xml')
    )

##############################################

# Because FieldInitParameters aren't normal RegisteredParameters, the
# classes have to be explicitly added to the xml docs.
xmlmenudump.addRegisteredClass(ScalarFieldInit)
xmlmenudump.addRegisteredClass(TwoVectorFieldInit)
xmlmenudump.addRegisteredClass(ThreeVectorFieldInit)

##############################################

# ResetFieldInit is used in Solve.Reset, to zero all Fields that don't
# have explicit initializers.

class ResetFieldInit (FieldInit):
    def func(self, position, index):
        return 0
